@@ -23,9 +23,9 @@ import os
 from rasterio.io import MemoryFile
 from requests_toolbelt import MultipartEncoder
 import warnings
+import threading
 
-
-__version__ = '1.2.6'
+__version__ = '1.2.7'
 url = 'https://api.ellipsis-drive.com/v1'
 
 s = requests.Session()
@@ -383,11 +383,11 @@ def geometryEdit(shapeId, layerId, geometryIds, features, token, zoomlevels = No
         features_sub =features_sub.to_json(na='drop')
         features_sub = json.loads(features_sub)
         
-        if str(type(zoomlevels)) == str(type(None)):
+        if str(type(zoomlevels)) != str(type(None)):
             zoomlevels = list(zoomlevels)
-            changes = [{'geometryId':x[0] , 'newProperties':x[1]['properties'], 'newGeometry':x[1]['geometry']} for x in zip(geometryIds_sub, features_sub['features'])]
-        else:
             changes = [{'geometryId':x[0] , 'newProperties':x[1]['properties'], 'newGeometry':x[1]['geometry'], 'newZoomlevels':zoomlevels} for x in zip(geometryIds_sub, features_sub['features'])]
+        else:
+            changes = [{'geometryId':x[0] , 'newProperties':x[1]['properties'], 'newGeometry':x[1]['geometry']} for x in zip(geometryIds_sub, features_sub['features'])]
             
         body = {"mapId":  mapId, "layerId":layerId, 'changes':changes}
         retried = 0
@@ -406,7 +406,7 @@ def geometryEdit(shapeId, layerId, geometryIds, features, token, zoomlevels = No
             raise ValueError(r.text)
 
 
-        loadingBar(i*3000 + len(indices_sub),features.shape[0])
+        loadingBar(i*1000 + len(indices_sub),features.shape[0])
 
 
 def geometryMove(shapeId, geometryIds, oldLayerId, newLayerId, token):
@@ -512,7 +512,7 @@ def geometryAdd(shapeId, layerId, features, token, zoomlevels=None):
         body = {"mapId":  mapId, "layerId":layerId, "features":features_sub['features']}
 
         if str(type(None)) != str(type(zoomlevels)):
-            body['zoomlevels'] = zoomlevels
+            body['zoomLevels'] = zoomlevels
 
         retried = 0
         while retried <= 20:
@@ -538,7 +538,7 @@ def geometryAdd(shapeId, layerId, features, token, zoomlevels=None):
 
         
     
-def messageGet(shapeId, layerId, userId= None, messageIds =None, geometryIds=None, limit = 100, deleted=False, token = None):
+def messageGet(shapeId, layerId, userId= None, messageIds =None, geometryIds=None, limit = None, deleted=False, token = None):
 
     body={'mapId':shapeId, 'layerId':layerId, 'deleted':deleted, 'returnType':'all'}
 
@@ -550,7 +550,10 @@ def messageGet(shapeId, layerId, userId= None, messageIds =None, geometryIds=Non
     keepGoing = True
     messages = []
     while(keepGoing):
-        body['pageSize'] = min(100, limit - len(messages))
+        if str(type(limit)) == str(type(None)):
+            body['pageSize'] = 100
+        else:
+            body['pageSize'] = min(100, limit - len(messages))
 
         if token == None:
             r = s.post(url + '/message/get',
@@ -566,12 +569,15 @@ def messageGet(shapeId, layerId, userId= None, messageIds =None, geometryIds=Non
         body['pageStart']=r['nextPageStart']
 
         messages = messages +  r['result']
-        keepGoing = len(messages) < limit or str(type(r['nextPageStart'])) == str(type(None))
+        if str(type(limit)) == str(type(None)):
+            keepGoing = str(type(r['nextPageStart'])) != str(type(None))
+        else:
+            keepGoing = len(messages) < limit or str(type(r['nextPageStart'])) != str(type(None))
         loadingBar(len(messages), len(messages))
     
     return(messages)
     
-def messageAdd(shapeId, layerId, geometryId, token, attachFile=None, replyTo = None, message = None, private= None, image=None, lon=None, lat=None): 
+def messageAdd(shapeId, layerId, geometryId, token, replyTo = None, message = None, private= None, image=None, lon=None, lat=None): 
     mapId = shapeId    
     body = {'mapId':mapId, 'layerId':layerId, 'geometryId':geometryId}
     if str(type(lon)) != "<class 'NoneType'>":
@@ -603,16 +609,6 @@ def messageAdd(shapeId, layerId, geometryId, token, attachFile=None, replyTo = N
     
     messageId = r.json()['id']
 
-    if str(type(attachFile)) != str(type(None)):
-            fileName = attachFile.split('/')[-1].split('/')[-1]
-
-            files = {'upload': open(attachFile,'rb')}
-            body = {'mapId':mapId, 'layerId':layerId, 'messageId':messageId, 'fileName':fileName}
-            r = s.post(url + '/message/attach', headers = {"Authorization":token},
-                         data = body, files =files )
-            if int(str(r).split('[')[1].split(']')[0]) != 200:
-                raise ValueError(r.text)
-
 
     return({'id':messageId})
 
@@ -645,19 +641,6 @@ def messageImage(shapeId, layerId, messageId, token = None):
  
     return(im)
 
-def messageAttachement(shapeId, layerId, messageId, fileName, token = None):
-    mapId = shapeId    
-    body = {'mapId':mapId, 'layerId':layerId, 'messageId':messageId}
-    if token ==None:
-        r = s.post(url + '/message/downloadAttachment',
-                     json = body )        
-    else:
-        r = s.post(url + '/message/downloadAttachment', headers = {"Authorization":token},
-                     json = body )
-    if int(str(r).split('[')[1].split(']')[0]) != 200:
-        raise ValueError(r.text)
-        
-    open(fileName, 'wb').write(r.content)
 
 
 def rasterAggregated(mapId, timestamp, geometry, approximate = True, token = None):
@@ -677,7 +660,7 @@ def rasterAggregated(mapId, timestamp, geometry, approximate = True, token = Non
     return(r.json())
 
 
-def rasterRaw(mapId, timestamp, xMin= None,xMax= None,yMin=None, yMax=None, zoom = None, num_bands = None, bounds=None, downsample = False, width = None, height = None, token = None):
+def rasterRaw(mapId, timestamp, xMin= None,xMax= None,yMin=None, yMax=None, zoom = None, num_bands = None, bounds=None, downsample = False, width = None, height = None, threads = 1, token = None):
     dtype = 'float32'
 
     if str(type(bounds)) != str(type(None)):
@@ -751,31 +734,50 @@ def rasterRaw(mapId, timestamp, xMin= None,xMax= None,yMin=None, yMax=None, zoom
             y_tiles = np.arange(min_y_osm, max_y_osm +1)
 
             r_total = np.zeros((256*(max_y_osm - min_y_osm + 1) ,256*(max_x_osm - min_x_osm + 1),num_bands), dtype = dtype)
-            
+
+            tiles = []            
             for tileY in y_tiles:
                 for tileX in x_tiles:
-                    x_index = tileX - min_x_osm
-                    y_index = tileY - min_y_osm
+                    tiles = tiles + [(tileX, tileY)]
 
-                    r = s.get(url + '/tileService/' + mapId + '/' + str(timestamp) + '/data/' + str(zoom) + '/' + str(tileX) + '/' + str(tileY) + token_inurl ,
-                                 timeout = 10 )
+            def subTiles(tiles):
+                    N = 0
+                    for tile in tiles:
+                        tileX = tile[0]
+                        tileY = tile[1]
+                        x_index = tileX - min_x_osm
+                        y_index = tileY - min_y_osm
+    
+                        r = s.get(url + '/tileService/' + mapId + '/' + str(timestamp) + '/data/' + str(zoom) + '/' + str(tileX) + '/' + str(tileY) + token_inurl ,
+                                     timeout = 10 )
+    
+    
+                        if int(str(r).split('[')[1].split(']')[0]) == 403:
+                                raise ValueError('insufficient access')
+                        if int(str(r).split('[')[1].split(']')[0]) != 200:
+                                r = np.zeros((num_bands,256,256))
+                        else:
+                                 with MemoryFile(r.content) as memfile:
+                                     with memfile.open() as dataset:
+                                         r = dataset.read()
+                        r = np.transpose(r, (1,2,0) )
+                        
+                        bands = list(np.arange(num_bands-1))
+                        bands.append(r.shape[2]-1)
+                        r_total[y_index*256:(y_index+1)*256,x_index*256:(x_index+1)*256, : ] = r[:,:, bands]
+                        loadingBar(N, len(tiles))
+                        N = N + 1
+                    
 
-
-                    if int(str(r).split('[')[1].split(']')[0]) != 200:
-                            r = np.zeros((num_bands,256,256))
-                    else:
-                             with MemoryFile(r.content) as memfile:
-                                 with memfile.open() as dataset:
-                                     r = dataset.read()
-                    r = np.transpose(r, (1,2,0) )
-                    
-                    bands = list(np.arange(num_bands-1))
-                    bands.append(r.shape[2]-1)
-                    r_total[y_index*256:(y_index+1)*256,x_index*256:(x_index+1)*256, : ] = r[:,:, bands]
-                    
-                    
-                    loadingBar(x_index + y_index*len(x_tiles) + 1,len(x_tiles)*len(y_tiles))
-                
+            size = math.floor(len(tiles)/threads) + 1
+            tiles_chunks = chunks(tiles, size)
+            prs = []
+            for tiles in tiles_chunks:
+                pr = threading.Thread(target = subTiles, args =(tiles,), daemon=True)
+                pr.start()
+                prs = prs + [pr] 
+            for pr in prs:
+                pr.join()                           
                 
             min_x_index = int(round((min_x_osm_precise - min_x_osm)*256))
             max_x_index = int(round((max_x_osm_precise- min_x_osm)*256))
@@ -798,7 +800,7 @@ def rasterRaw(mapId, timestamp, xMin= None,xMax= None,yMin=None, yMax=None, zoom
 
 
 
-def rasterVisual(mapId, timestamp, layerId, xMin= None,xMax= None,yMin=None, yMax=None, bounds=None, downsample = False, width = None, height=None, token = None):
+def rasterVisual(mapId, timestamp, layerId, xMin= None,xMax= None,yMin=None, yMax=None, bounds=None, downsample = False, width = None, height=None, threads = 1, token = None):
 
     dtype = 'uint8'
 
@@ -860,10 +862,19 @@ def rasterVisual(mapId, timestamp, layerId, xMin= None,xMax= None,yMin=None, yMa
             x_tiles = np.arange(min_x_osm, max_x_osm+1)
             y_tiles = np.arange(min_y_osm, max_y_osm +1)
 
-            r_total = np.zeros((256*(max_y_osm - min_y_osm + 1) ,256*(max_x_osm - min_x_osm + 1),4), dtype = dtype)
-            
+
+            tiles = []            
             for tileY in y_tiles:
                 for tileX in x_tiles:
+                    tiles = tiles + [(tileX,tileY)]
+                    
+                    
+            def subTiles(tiles):
+                N=0
+                for tile in tiles:
+                    tileX = tile[0]
+                    tileY = tile[1]
+                    
                     x_index = tileX - min_x_osm
                     y_index = tileY - min_y_osm
 
@@ -879,9 +890,21 @@ def rasterVisual(mapId, timestamp, layerId, xMin= None,xMax= None,yMin=None, yMa
 
                     r_total[y_index*256:(y_index+1)*256,x_index*256:(x_index+1)*256, : ] = r
                         
+                    loadingBar(N, len(tiles))
+                    N = N+1
 
-                    loadingBar(x_index + y_index*len(x_tiles) + 1,len(x_tiles)*len(y_tiles))
-                
+
+            r_total = np.zeros((256*(max_y_osm - min_y_osm + 1) ,256*(max_x_osm - min_x_osm + 1),4), dtype = dtype)
+
+            size = math.floor(len(tiles)/threads) + 1
+            tiles_chunks = chunks(tiles, size)
+            prs = []
+            for tiles in tiles_chunks:
+                pr = threading.Thread(target = subTiles, args =(tiles,), daemon=True)
+                pr.start()
+                prs = prs + [pr] 
+            for pr in prs:
+                pr.join()                     
                 
             min_x_index = int(round((min_x_osm_precise - min_x_osm)*256))
             max_x_index = int(round((max_x_osm_precise- min_x_osm)*256))
@@ -1471,7 +1494,7 @@ def chunks(l, n = 3000):
         result.append(l[i:i+n])
     return(result)
     
- 
+
  
 def cover(bounds, w):
     if str(type(bounds)) == "<class 'shapely.geometry.polygon.Polygon'>" :
